"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GptEncoding = exports.DISALLOW_ALL_SPECIAL_TOKENS = void 0;
/* eslint-disable no-param-reassign */
const BytePairEncodingCore_js_1 = require("./BytePairEncodingCore.js");
const escapeRegExp_js_1 = require("./escapeRegExp.js");
const mapping_js_1 = require("./mapping.js");
const modelParams_js_1 = require("./modelParams.js");
const specialTokens_js_1 = require("./specialTokens.js");
const utfUtil_js_1 = require("./utfUtil.js");
exports.DISALLOW_ALL_SPECIAL_TOKENS = 'all';
class GptEncoding {
    static EndOfPrompt = specialTokens_js_1.EndOfPrompt;
    static EndOfText = specialTokens_js_1.EndOfText;
    static FimMiddle = specialTokens_js_1.FimMiddle;
    static FimPrefix = specialTokens_js_1.FimPrefix;
    static FimSuffix = specialTokens_js_1.FimSuffix;
    decoder = new TextDecoder('utf8');
    bytePairEncodingCoreProcessor;
    specialTokenMapping;
    constructor({ tokenSplitRegex, mergeableBytePairRanks, specialTokenMapping, expectedVocabularySize, }) {
        const maxTokenValue = Math.max(GptEncoding.getMaxValueFromMap(mergeableBytePairRanks), GptEncoding.getMaxValueFromMap(specialTokenMapping));
        this.specialTokenMapping = specialTokenMapping;
        if (expectedVocabularySize !== undefined) {
            if (mergeableBytePairRanks.size + specialTokenMapping.size !==
                expectedVocabularySize) {
                throw new Error('The number of mergeable tokens and special tokens must be equal to explicit_n_vocab.');
            }
            if (maxTokenValue !== expectedVocabularySize - 1) {
                throw new Error('The maximum token value must be equal to explicit_n_vocab - 1.');
            }
        }
        this.bytePairEncodingCoreProcessor = new BytePairEncodingCore_js_1.BytePairEncodingCore({
            bytePairEncoder: mergeableBytePairRanks,
            specialTokenEncoder: specialTokenMapping,
            tokenSplitRegex,
        });
        this.encode = this.encode.bind(this);
        this.decode = this.decode.bind(this);
        this.encodeGenerator = this.encodeGenerator.bind(this);
        this.decodeGenerator = this.decodeGenerator.bind(this);
        this.decodeAsyncGenerator = this.decodeAsyncGenerator.bind(this);
        this.isWithinTokenLimit = this.isWithinTokenLimit.bind(this);
    }
    static getEncodingApi(encodingName, getMergeableRanks) {
        const modelParams = (0, modelParams_js_1.getModelParams)(encodingName, getMergeableRanks);
        return new GptEncoding(modelParams);
    }
    static getEncodingApiForModel(modelName, getMergeableRanks) {
        const encodingName = mapping_js_1.modelToEncodingMap[modelName];
        return GptEncoding.getEncodingApi(encodingName, getMergeableRanks);
    }
    static async getEncodingApiAsync(encodingName, getMergeableRanks) {
        const modelParams = await (0, modelParams_js_1.getModelParamsAsync)(encodingName, getMergeableRanks);
        return new GptEncoding(modelParams);
    }
    static async getEncodingApiForModelAsync(modelName, getMergeableRanks) {
        const encodingName = mapping_js_1.modelToEncodingMap[modelName];
        return GptEncoding.getEncodingApiAsync(encodingName, getMergeableRanks);
    }
    static getMaxValueFromMap(map) {
        let max = 0;
        map.forEach((val) => {
            max = Math.max(max, val);
        });
        return max;
    }
    static specialTokenRegex(tokens) {
        const escapedTokens = [...tokens].map(escapeRegExp_js_1.escapeRegExp);
        const inner = escapedTokens.join('|');
        return new RegExp(`(${inner})`);
    }
    encodeGenerator(lineToEncode, allowedSpecial = new Set(), disallowedSpecial = new Set([
        exports.DISALLOW_ALL_SPECIAL_TOKENS,
    ])) {
        const specialTokensSet = new Set(this.specialTokenMapping.keys());
        if (disallowedSpecial.has(exports.DISALLOW_ALL_SPECIAL_TOKENS)) {
            disallowedSpecial = new Set(specialTokensSet);
            disallowedSpecial.forEach((val) => allowedSpecial.delete(val));
        }
        if (allowedSpecial.has(exports.DISALLOW_ALL_SPECIAL_TOKENS)) {
            allowedSpecial = specialTokensSet;
        }
        if (disallowedSpecial.size > 0) {
            const regexPattern = GptEncoding.specialTokenRegex(disallowedSpecial);
            const match = lineToEncode.match(regexPattern);
            if (match !== null) {
                throw new Error(`Disallowed special token found: ${match[0]}`);
            }
        }
        return this.bytePairEncodingCoreProcessor.encodeNative(lineToEncode, allowedSpecial);
    }
    encode(lineToEncode, allowedSpecial = new Set(), disallowedSpecial = new Set([
        exports.DISALLOW_ALL_SPECIAL_TOKENS,
    ])) {
        return [
            ...this.encodeGenerator(lineToEncode, allowedSpecial, disallowedSpecial),
        ].flat();
    }
    /**
     * @returns {false | number} false if token limit is exceeded, otherwise the number of tokens
     */
    isWithinTokenLimit(text, tokenLimit) {
        const tokenGenerator = this.encodeGenerator(text);
        let count = 0;
        for (const tokens of tokenGenerator) {
            count += tokens.length;
            if (count > tokenLimit) {
                return false;
            }
        }
        return count;
    }
    *decodeGenerator(inputTokensToDecode) {
        const decodedByteGenerator = this.bytePairEncodingCoreProcessor.decodeNative(inputTokensToDecode);
        let buffer = '';
        for (const decodedPart of decodedByteGenerator) {
            buffer += this.decoder.decode(decodedPart, { stream: true });
            if (buffer.length === 0 || (0, utfUtil_js_1.endsWithIncompleteUtfPairSurrogate)(buffer)) {
                // Keep the high surrogate in the buffer and continue with the next token
                // eslint-disable-next-line no-continue
                continue;
            }
            else {
                yield buffer;
                // reset buffer
                buffer = '';
            }
        }
        // Yield any remaining characters in the buffer
        if (buffer.length > 0) {
            yield buffer;
        }
    }
    async *decodeAsyncGenerator(inputTokensToDecode) {
        const decodedByteGenerator = this.bytePairEncodingCoreProcessor.decodeNativeAsync(inputTokensToDecode);
        let buffer = '';
        for await (const decodedPart of decodedByteGenerator) {
            buffer += this.decoder.decode(decodedPart, { stream: true });
            if (buffer.length === 0 || (0, utfUtil_js_1.endsWithIncompleteUtfPairSurrogate)(buffer)) {
                // Keep the high surrogate in the buffer and continue with the next token
                // eslint-disable-next-line no-continue
                continue;
            }
            else {
                yield buffer;
                // reset buffer
                buffer = '';
            }
        }
        // Yield any remaining characters in the buffer
        if (buffer.length > 0) {
            yield buffer;
        }
    }
    decode(inputTokensToDecode) {
        return [...this.decodeGenerator(inputTokensToDecode)].join('');
    }
}
exports.GptEncoding = GptEncoding;
//# sourceMappingURL=GptEncoding.js.map